import json
import os.path

import aiohttp
from aiocache import cached

from .exception import *
from .model import EnkaData
from .model.artifact import Artifact
from .model.character import CharacterSkill, CharacterConstellation, CharacterSkillType


class Enka:
    """
    Main Enka object.

    Example::

        user = await client.fetch_user(104267816)
        print(f"Nickname: {user.player.nickname}")

    """
    _URL = "https://enka.network/api/uid/{uid}"
    # https://github.com/theBowja/GenshinData-1
    # https://raw.githubusercontent.com/GrownNed/Homework/master
    _REPO_BASE = 'https://gitlab.com/Dimbreath/gamedata/-/raw/master'
    _LANG_URL = _REPO_BASE + '/TextMap/TextMap{lang}.json?inline=false'
    _AVATAR_URL = _REPO_BASE + '/ExcelBinOutput/AvatarExcelConfigData.json?inline=false'
    _TALENT_URL = _REPO_BASE + '/ExcelBinOutput/AvatarTalentExcelConfigData.json?inline=false'
    _SKILL_DEPOT_URL = _REPO_BASE + '/ExcelBinOutput/AvatarSkillDepotExcelConfigData.json?inline=false'
    _SKILL_URL = _REPO_BASE + '/ExcelBinOutput/AvatarSkillExcelConfigData.json?inline=false'
    USER_AGENT = "Mozilla/5.0"
    timeout = 30
    """Http connection timeout"""
    proxy = ''
    """Http connection proxy"""
    _lang_data = {}
    """Internal language data"""
    _avatar_data = {}
    """Internal avatar data"""
    _talent_data = {}
    """Internal talent data"""
    _skill_depot_data = {}
    """Internal skill pot data"""
    _skill_data = {}
    """Internal skill data"""
    lang = 'en'
    """Language for text hash resolve"""
    _cache = 'cache'
    """Cache folder name"""

    def _cache_exists(self, name):
        """
        Check if cache file exists and return full cache file path

        :param name: cache name
        :return: full cache file path
        """
        package_path = os.path.dirname(os.path.realpath(__file__))
        cache_path = os.path.join(package_path, self._cache)
        cache_file_path = os.path.join(cache_path, name)

        if not os.path.exists(cache_path):
            try:
                os.mkdir(cache_path)
            except OSError:
                pass
            return ''
        if not os.path.isdir(cache_path):
            try:
                os.remove(cache_path)
            except OSError:
                pass
            return ''
        if os.path.exists(cache_file_path):
            return cache_file_path
        else:
            return ''

    async def _process_cache(self, client: aiohttp.ClientSession, url, cache_file, force_cache=False,
                             force_update=False):
        """
        Process cache

        :param client: aiohttp client
        :param url: git url
        :param cache_file: cache file name
        :param force_cache: ignore git update, force using cache file
        :param force_update: force update from git repo
        :return: cache file path
        """
        cache_file_path = self._cache_exists(cache_file)
        if cache_file_path and not force_update:
            stat = os.stat(cache_file_path)
            file_size = stat.st_size
            if not force_cache:
                resp = await client.head(url, proxy=self.proxy)
                cl = int(resp.headers['Content-Length'])
                if cl != file_size:
                    resp = await client.get(url, proxy=self.proxy)
                    with open(cache_file_path, 'wb') as f:
                        f.write(await resp.read())
        else:
            package_path = os.path.dirname(os.path.realpath(__file__))
            cache_path = os.path.join(package_path, self._cache)
            cache_file_path = os.path.join(cache_path, cache_file)
            resp = await client.get(url, proxy=self.proxy)
            with open(cache_file_path, 'wb') as f:
                f.write(await resp.read())
        return cache_file_path

    async def load_lang(self, lang='en', force_cache=False, force_update=False):
        """
        Load language data from cache or Dimbreath repo

        :param lang: language you want to load, default 'en'
        :param force_cache: Do not check repo update if cache file exists
        :param force_update: force update from git repo
        """

        async with aiohttp.ClientSession(headers={"User-Agent": self.USER_AGENT}) as client:
            if lang not in self._lang_data:
                cache_file_path = await self._process_cache(client,
                                                            self._LANG_URL.format(lang=lang.upper()),
                                                            'lang.json',
                                                            force_cache,
                                                            force_update)
                with open(cache_file_path, 'rb') as f:
                    self._lang_data[lang] = json.load(f)
            if not self._avatar_data:
                cache_file_path = await self._process_cache(client,
                                                            self._AVATAR_URL,
                                                            'avatar.json',
                                                            force_cache,
                                                            force_update)
                with open(cache_file_path, 'rb') as f:
                    for x in json.load(f):
                        self._avatar_data[x['id']] = x
            if not self._skill_depot_data:
                cache_file_path = await self._process_cache(client,
                                                            self._SKILL_DEPOT_URL,
                                                            'skill_depot.json',
                                                            force_cache,
                                                            force_update)
                with open(cache_file_path, 'rb') as f:
                    for x in json.load(f):
                        self._skill_depot_data[x['id']] = x
            if not self._skill_data:
                cache_file_path = await self._process_cache(client,
                                                            self._SKILL_URL,
                                                            'skill.json',
                                                            force_cache,
                                                            force_update)
                with open(cache_file_path, 'rb') as f:
                    for x in json.load(f):
                        self._skill_data[x['id']] = x
            if not self._talent_data:
                cache_file_path = await self._process_cache(client,
                                                            self._TALENT_URL,
                                                            'talent.json',
                                                            force_cache,
                                                            force_update)
                with open(cache_file_path, 'rb') as f:
                    for x in json.load(f):
                        self._talent_data[x['talentId']] = x

    async def resolve_text_hash(self, text_hash, lang='en'):
        """
        Resolve text hash to actual text

        :param text_hash: text hash
        :param lang: language you want to resolve to
        :return: resolved text
        """
        if lang not in self._lang_data:
            await self.load_lang(lang)
        if not isinstance(text_hash, str):
            text_hash = str(text_hash)
        if text_hash in self._lang_data[lang]:
            return self._lang_data[lang][text_hash]
        else:
            return ''

    @cached(ttl=600)
    async def fetch_user(self, uid: int, player_only=False) -> EnkaData:
        """
        Fetch user data from enka api, resolve text hash if available

        :param uid: user in game uid
        :param player_only: fetch only player info
        :return: EnkaData object referencing player
        """
        if not isinstance(uid, int):
            try:
                uid = int(uid)
            except ValueError:
                raise ValidateUIDError("Validate UID failed. Please check your UID.")
        if len(str(uid)) != 9 or (100000000 > uid > 999999999):
            raise ValidateUIDError("Validate UID failed. Please check your UID.")

        async with aiohttp.ClientSession(headers={"User-Agent": self.USER_AGENT},
                                         timeout=aiohttp.ClientTimeout(total=self.timeout)) as client:
            url = self._URL.format(uid=uid)
            if player_only:
                url += '?info'
            resp = await client.get(url, proxy=self.proxy)

            if resp.status != 200:
                if resp.status == 400:
                    raise WrongUIDFormat(f'Wrong uid format! {uid}')
                elif resp.status == 404:
                    raise PlayerDoesNOTExist(f'Player {uid} does not exist')
                elif resp.status == 424:
                    raise GameMaintenance('Game in maintenance')
                elif resp.status == 429:
                    raise RateLimited('Rate limited')
                elif resp.status == 500:
                    raise GeneralServerError('Server error 500')
                elif resp.status == 503:
                    raise EnkaError('Enka endpoint error 503')
                else:
                    raise EnkaError(f'Unknown error {resp.status}')

            data = await resp.json()

            if not data:
                raise EnkaError(f'Unknown error[No Data]')

        obj: EnkaData = EnkaData.parse_obj(data)

        if not player_only:
            for character in obj.characters:
                if character.skill_depot_id in self._skill_depot_data:
                    depot = self._skill_depot_data[character.skill_depot_id]
                    if 'energySkill' in depot:
                        burst_id = depot['energySkill']
                        if burst_id in self._skill_data:
                            cs = CharacterSkill()
                            cs.id = burst_id
                            cs.type = CharacterSkillType.ElementalBurst
                            cs.name_hash = self._skill_data[burst_id]['nameTextMapHash']
                            cs.icon = self._skill_data[burst_id]['skillIcon']
                            character.skills.append(cs)
                    for skill_id in depot['skills']:
                        if skill_id and skill_id in self._skill_data:
                            skill_info = self._skill_data[skill_id]
                            cs = CharacterSkill()
                            cs.id = skill_id
                            if 'cdTime' in skill_info and skill_info['cdTime']:
                                cs.type = CharacterSkillType.ElementalSkill
                            else:
                                cs.type = CharacterSkillType.NormalSkill
                            cs.name_hash = skill_info['nameTextMapHash']
                            cs.icon = self._skill_data[burst_id]['skillIcon']
                            character.skills.append(cs)
                    for talent_id in depot['talents']:
                        if talent_id and talent_id in self._talent_data:
                            talent_info = self._talent_data[talent_id]
                            tl = CharacterConstellation()
                            tl.id = talent_id
                            tl.icon = talent_info['icon']
                            tl.name_hash = talent_info['nameTextMapHash']
                            character.constellations.append(tl)
                    character.process_skill()
                    character.activate_constellation()
                if self.lang and self.lang in self._lang_data and self._lang_data[self.lang] and self._avatar_data:
                    for equip in character.equipList:
                        equip.flat.nameText = await self.resolve_text_hash(equip.flat.nameTextMapHash, self.lang)
                        if isinstance(equip, Artifact):
                            equip.flat.setNameText = await self.resolve_text_hash(equip.flat.setNameTextMapHash,
                                                                                  self.lang)
                    if character.id in self._avatar_data:
                        character.name = await self.resolve_text_hash(
                            self._avatar_data[character.id]['nameTextMapHash'],
                            self.lang)
                    for skill in character.skills:
                        skill.name = await self.resolve_text_hash(skill.name_hash, self.lang)
                    for c in character.constellations:
                        c.name = await self.resolve_text_hash(c.name_hash, self.lang)
        return obj



from enum import Enum
from typing import List, Dict, Union, Optional

from pydantic import BaseModel, Field, validator

from .artifact import Artifact
from .combats import CharacterCombat
from .weapon import Weapon


class CharacterPropertyType(int, Enum):
    """ """
    UNKNOWN = -1
    XP = 1001
    ASCENSION = 1002
    LEVEL = 4001


class CharacterFriendshipLevel(BaseModel):
    """ """
    level: int = Field(0, alias="expLevel")


class CharacterProperty(BaseModel):
    """ """
    type: CharacterPropertyType
    ival: str = Field("", alias="ival")
    value: str = Field("", alias="val")

    @validator('type', pre=True)
    def check_type(cls, val: int):
        if val not in [1001, 1002, 4001]:
            val = -1

        return val

    class Config:
        """ """
        use_enum_values = True


class CharacterConstellation:
    name = ''
    icon = ''
    id = 0
    name_hash = 0
    activated = False


class CharacterSkillType(int, Enum):
    NormalSkill = 0
    ElementalSkill = 1
    ElementalBurst = 2


class CharacterSkill:
    name = ''
    level = 0
    icon = ''
    id = 0
    name_hash = 0
    type: CharacterSkillType


class CharacterInfo(BaseModel):
    """
    Character info class for every character shown in game
    """
    name: Optional[str] = Field('')
    """Character name"""
    id: int = Field(0, alias="avatarId")
    "Character avatar id for fetching more detail later"
    friendship: CharacterFriendshipLevel = Field({}, alias="fetterInfo")
    "Character friendship"
    properties: Dict[str, CharacterProperty] = Field({}, alias="propMap")
    "Some character properties like xp, ascension , level"
    # Artifacts
    equipList: List[Union[Artifact, Weapon]] = Field([], alias="equipList")
    "Everything equipped on character, including artifacts and weapon"
    combat: CharacterCombat = Field({}, alias="fightPropMap")
    "Character combat stats"
    inherentProudSkillList: List[int] = Field([], alias="inherentProudSkillList")
    "Character skill data"
    skill_depot_id: int = Field(0, alias="skillDepotId")
    """Character skill depot id"""
    internal_constellations: List[int] = Field([], alias='talentIdList')
    """Internal Character talents"""
    constellations: Optional[List[CharacterConstellation]] = Field([])
    """Character talents"""
    skill_level: Dict[int, int] = Field({}, alias="skillLevelMap")
    """Character skill level map, skill_id:skill_level"""
    skills: Optional[List[CharacterSkill]] = Field([])
    """Character skill info"""

    @property
    def ascension(self):
        for key, val in self.properties.items():
            if int(key) == CharacterPropertyType.ASCENSION:
                return val.ival
        return 0

    @property
    def level(self):
        for key, val in self.properties.items():
            if int(key) == CharacterPropertyType.LEVEL:
                return val.ival
        return 0

    @property
    def experience(self):
        for key, val in self.properties.items():
            if int(key) == CharacterPropertyType.XP:
                return val.ival
        return 0

    @property
    def artifacts(self) -> List[Artifact]:
        """
        :return: Artifact lists equipped on character
        """
        artifacts = []
        for _ in self.equipList:
            if isinstance(_, Artifact):
                artifacts.append(_)
        return artifacts

    @property
    def weapon(self) -> Weapon:
        """
        :return: weapon equipped on character
        """
        for _ in self.equipList:
            if isinstance(_, Weapon):
                return _

    def activate_constellation(self):
        for c in self.constellations:
            if c.id in self.internal_constellations:
                c.activated = True

    def process_skill(self):
        for s in self.skills:
            if s.id in self.skill_level:
                s.level = self.skill_level[s.id]

    class Config:
        arbitrary_types_allowed = True



from typing import List, Optional

from pydantic import BaseModel


class ProfilePicture(BaseModel):
    avatarId: Optional[int]


class ShowAvatar(BaseModel):
    avatarId: str
    """Character id"""
    level: int
    """Character level"""


class PlayerInfo(BaseModel):
    """
    Player info class
    """
    finishAchievementNum: Optional[int]
    """Total achievement"""
    level: int
    """Player level"""
    nameCardId: int
    """Player name card background id"""
    nickname: Optional[str]
    """Player nickname"""
    signature: Optional[str]
    """Player signature"""
    worldLevel: Optional[int]
    """Player world level"""
    profilePicture: ProfilePicture
    """Player profile picture"""

    showAvatarInfoList: Optional[List[ShowAvatar]]
    """Player characters shown"""

    towerFloorIndex: Optional[int]
    """Player abyss floor"""
    towerLevelIndex: Optional[int]
    """Player abyss level"""



from pydantic import BaseModel, Field


class CharacterCombat(BaseModel):
    BASE_HP: int = Field(0, alias="1")
    FIGHT_PROP_HP: int = Field(0, alias="2")
    FIGHT_PROP_HP_PERCENT: float = Field(0, alias="3")
    FIGHT_PROP_BASE_ATTACK: int = Field(0, alias="4")
    FIGHT_PROP_ATTACK: int = Field(0, alias="5")
    FIGHT_PROP_ATTACK_PERCENT: float = Field(0, alias="6")
    FIGHT_PROP_BASE_DEFENSE: int = Field(0, alias="7")
    FIGHT_PROP_DEFENSE: int = Field(0, alias="8")
    FIGHT_PROP_DEFENSE_PERCENT: float = Field(0, alias="9")
    FIGHT_PROP_BASE_SPEED: int = Field(0, alias="10")
    FIGHT_PROP_SPEED_PERCENT: float = Field(0, alias="11")
    FIGHT_PROP_HP_MP_PERCENT: float = Field(0, alias="12")
    FIGHT_PROP_ATTACK_MP_PERCENT: float = Field(0, alias="13")
    FIGHT_PROP_CRITICAL: float = Field(0, alias="20")
    FIGHT_PROP_ANTI_CRITICAL: float = Field(0, alias="21")
    FIGHT_PROP_CRITICAL_HURT: float = Field(0, alias="22")
    FIGHT_PROP_CHARGE_EFFICIENCY: float = Field(0, alias="23")
    FIGHT_PROP_ADD_HURT: float = Field(0, alias="24")
    FIGHT_PROP_SUB_HURT: float = Field(0, alias="25")
    FIGHT_PROP_HEAL_ADD: float = Field(0, alias="26")
    FIGHT_PROP_HEALED_ADD: float = Field(0, alias="27")
    FIGHT_PROP_ELEMENT_MASTERY: int = Field(0, alias="28")
    FIGHT_PROP_PHYSICAL_SUB_HURT: float = Field(0, alias="29")
    FIGHT_PROP_PHYSICAL_ADD_HURT: float = Field(0, alias="30")
    FIGHT_PROP_DEFENCE_IGNORE_RATIO: float = Field(0, alias="31")
    FIGHT_PROP_DEFENCE_IGNORE_DELTA: float = Field(0, alias="32")
    FIGHT_PROP_FIRE_ADD_HURT: float = Field(0, alias="40")
    FIGHT_PROP_ELEC_ADD_HURT: float = Field(0, alias="41")
    FIGHT_PROP_WATER_ADD_HURT: float = Field(0, alias="42")
    FIGHT_PROP_GRASS_ADD_HURT: float = Field(0, alias="43")
    FIGHT_PROP_WIND_ADD_HURT: float = Field(0, alias="44")
    FIGHT_PROP_ROCK_ADD_HURT: float = Field(0, alias="45")
    FIGHT_PROP_ICE_ADD_HURT: float = Field(0, alias="46")
    FIGHT_PROP_HIT_HEAD_ADD_HURT: float = Field(0, alias="47")
    FIGHT_PROP_FIRE_SUB_HURT: float = Field(0, alias="50")
    FIGHT_PROP_ELEC_SUB_HURT: float = Field(0, alias="51")
    FIGHT_PROP_WATER_SUB_HURT: float = Field(0, alias="52")
    FIGHT_PROP_GRASS_SUB_HURT: float = Field(0, alias="53")
    FIGHT_PROP_WIND_SUB_HURT: float = Field(0, alias="54")
    FIGHT_PROP_ROCK_SUB_HURT: float = Field(0, alias="55")
    FIGHT_PROP_ICE_SUB_HURT: float = Field(0, alias="56")
    FIGHT_PROP_EFFECT_HIT: float = Field(0, alias="60")
    FIGHT_PROP_EFFECT_RESIST: float = Field(0, alias="61")
    FIGHT_PROP_FREEZE_RESIST: float = Field(0, alias="62")
    FIGHT_PROP_TORPOR_RESIST: float = Field(0, alias="63")
    FIGHT_PROP_DIZZY_RESIST: float = Field(0, alias="64")
    FIGHT_PROP_FREEZE_SHORTEN: float = Field(0, alias="65")
    FIGHT_PROP_TORPOR_SHORTEN: float = Field(0, alias="66")
    FIGHT_PROP_DIZZY_SHORTEN: float = Field(0, alias="67")
    FIGHT_PROP_MAX_FIRE_ENERGY: int = Field(0, alias="70")
    FIGHT_PROP_MAX_ELEC_ENERGY: int = Field(0, alias="71")
    FIGHT_PROP_MAX_WATER_ENERGY: int = Field(0, alias="72")
    FIGHT_PROP_MAX_GRASS_ENERGY: int = Field(0, alias="73")
    FIGHT_PROP_MAX_WIND_ENERGY: int = Field(0, alias="74")
    FIGHT_PROP_MAX_ICE_ENERGY: int = Field(0, alias="75")
    FIGHT_PROP_MAX_ROCK_ENERGY: int = Field(0, alias="76")
    FIGHT_PROP_SKILL_CD_MINUS_RATIO: int = Field(0, alias="80")
    FIGHT_PROP_SHIELD_COST_MINUS_RATIO: int = Field(0, alias="81")

    FIGHT_PROP_CUR_FIRE_ENERGY: int = Field(0, alias="1000")
    FIGHT_PROP_CUR_ELEC_ENERGY: int = Field(0, alias="1001")
    FIGHT_PROP_CUR_WATER_ENERGY: int = Field(0, alias="1002")
    FIGHT_PROP_CUR_GRASS_ENERGY: int = Field(0, alias="1003")
    FIGHT_PROP_CUR_WIND_ENERGY: int = Field(0, alias="1004")
    FIGHT_PROP_CUR_ICE_ENERGY: int = Field(0, alias="1005")
    FIGHT_PROP_CUR_ROCK_ENERGY: int = Field(0, alias="1006")
    FIGHT_PROP_CUR_HP: int = Field(0, alias="1010")

    FIGHT_PROP_MAX_HP: int = Field(0, alias="2000")
    FIGHT_PROP_CUR_ATTACK: int = Field(0, alias="2001")
    FIGHT_PROP_CUR_DEFENSE: int = Field(0, alias="2002")
    FIGHT_PROP_CUR_SPEED: int = Field(0, alias="2003")

    FIGHT_PROP_NONEXTRA_ATTACK: int = Field(0, alias="3000")
    FIGHT_PROP_NONEXTRA_DEFENSE: int = Field(0, alias="3001")
    FIGHT_PROP_NONEXTRA_CRITICAL: int = Field(0, alias="3002")
    FIGHT_PROP_CUR_SPEED: int = Field(0, alias="3003")
    FIGHT_PROP_NONEXTRA_CRITICAL_HURT: int = Field(0, alias="3004")
    FIGHT_PROP_NONEXTRA_CHARGE_EFFICIENCY: int = Field(0, alias="3005")
    FIGHT_PROP_NONEXTRA_ELEMENT_MASTERY: int = Field(0, alias="3006")
    FIGHT_PROP_NONEXTRA_PHYSICAL_SUB_HURT: int = Field(0, alias="3007")
    FIGHT_PROP_NONEXTRA_FIRE_ADD_HURT: int = Field(0, alias="3008")
    FIGHT_PROP_NONEXTRA_ELEC_ADD_HURT: int = Field(0, alias="3009")
    FIGHT_PROP_NONEXTRA_WATER_ADD_HURT: int = Field(0, alias="3010")
    FIGHT_PROP_NONEXTRA_GRASS_ADD_HURT: int = Field(0, alias="3011")
    FIGHT_PROP_NONEXTRA_WIND_ADD_HURT: int = Field(0, alias="3012")
    FIGHT_PROP_NONEXTRA_ROCK_ADD_HURT: int = Field(0, alias="3013")
    FIGHT_PROP_NONEXTRA_ICE_ADD_HURT: int = Field(0, alias="3014")
    FIGHT_PROP_NONEXTRA_FIRE_SUB_HURT: int = Field(0, alias="3015")
    FIGHT_PROP_NONEXTRA_ELEC_SUB_HURT: int = Field(0, alias="3016")
    FIGHT_PROP_NONEXTRA_WATER_SUB_HURT: int = Field(0, alias="3017")
    FIGHT_PROP_NONEXTRA_GRASS_SUB_HURT: int = Field(0, alias="3018")
    FIGHT_PROP_NONEXTRA_WIND_SUB_HURT: int = Field(0, alias="3019")
    FIGHT_PROP_NONEXTRA_ROCK_SUB_HURT: int = Field(0, alias="3020")
    FIGHT_PROP_NONEXTRA_ICE_SUB_HURT: int = Field(0, alias="3021")
    FIGHT_PROP_NONEXTRA_SKILL_CD_MINUS_RATIO: int = Field(0, alias="3022")
    FIGHT_PROP_NONEXTRA_SHIELD_COST_MINUS_RATIO: int = Field(0, alias="3023")
    FIGHT_PROP_NONEXTRA_PHYSICAL_ADD_HURT: int = Field(0, alias="3024")

    RAW: dict = {}

    def __init__(self, **data):
        super().__init__(**data)
        self.RAW = data



from typing import Dict, List, Optional

from pydantic import BaseModel, Field


class WeaponInfo(BaseModel):
    level: int
    affixMap: Optional[Dict[int, int]]
    promoteLevel: Optional[int]


class WeaponStats(BaseModel):
    appendPropId: str
    statValue: int


class WeaponFlat(BaseModel):
    icon: str
    itemType: str
    nameTextMapHash: str
    nameText: Optional[str] = Field('')
    rankLevel: int
    weaponStats: List[WeaponStats]

    @property
    def name(self):
        return self.nameText

    @property
    def rank(self):
        return self.rankLevel


class Weapon(BaseModel):
    """Weapon info class"""
    id: int = Field(0, alias="itemId")
    """Weapon id"""
    data: WeaponInfo = Field({}, alias="weapon")
    """Weapon data"""
    flat: WeaponFlat = Field({})
    """Weapon flat data, anything inside flat can be access directly in Weapon class too."""

    @property
    def level(self) -> int:
        """Weapon level"""
        return self.data.level

    @property
    def refine(self) -> int:
        for key, val in self.data.affixMap.items():
            if val:
                return val + 1
        return 0

    def __getattr__(self, item):
        if hasattr(self.flat, item):
            return getattr(self.flat, item)
        else:
            return None



from typing import List, Optional

from pydantic import BaseModel, Field


class ArtifactProperty(BaseModel):
    mainPropId: str = Field('', alias='mainPropId')
    statValue: float = Field(0, alias='statValue')
    appendPropId: str = Field('', alias='appendPropId')

    @property
    def prop(self) -> str:
        """Artifact property"""
        if self.mainPropId:
            return self.mainPropId
        else:
            return self.appendPropId

    @property
    def value(self) -> float:
        """Artifact property value"""
        return self.statValue


class ArtifactFlat(BaseModel):
    equipType: str
    """Artifact equip types
    
    EQUIP_BRACER: flower
    
    EQUIP_NECKLACE: feather
    
    EQUIP_SHOES: sand
    
    EQUIP_RING: goblet
    
    EQUIP_DRESS: circlet"""
    icon: str
    """Artifact icon"""
    itemType: str
    """Artifact item type, mostly 'ITEM_RELIQUARY'"""
    nameTextMapHash: str
    """Artifact name hash"""
    nameText: Optional[str] = Field('')
    """Artifact name text
    
    *Note: You must called load_lang before otherwise this field is empty*"""
    rankLevel: int
    """Artifact rank, 5/4/3/2/1 star"""
    main_stat: ArtifactProperty = Field({}, alias="reliquaryMainstat")
    """Artifact main stats"""
    sub_stats: List[ArtifactProperty] = Field({}, alias="reliquarySubstats")
    """Artifact sub stats"""
    setNameTextMapHash: str
    """Artifact name hash"""
    setNameText: Optional[str] = Field('')
    """Artifact set name text
    
    *Note: You must called load_lang before otherwise this field is empty*"""

    @property
    def icon_url(self) -> str:
        """
        :return: Artifact icon url from https://enka.shinshin.moe/ui/
        """
        return f'https://enka.network/ui/{self.icon}.png'

    @property
    def name(self):
        return self.nameText

    @property
    def set_name(self):
        return self.setNameText


class ArtifactInfo(BaseModel):
    level: int
    mainPropId: int
    appendPropIdList: Optional[List[int]] = Field([])


class Artifact(BaseModel):
    """Artifact information class"""
    id: int = Field(0, alias="itemId")
    """Artifact id"""
    data: ArtifactInfo = Field({}, alias="reliquary")
    """Some Artifact info like upgrade level, upgrade value etc."""
    flat: ArtifactFlat
    """Artifact flat info class, anything inside flat can be access directly in Artifact class too."""

    @property
    def level(self) -> int:
        """Artifact level"""
        return self.data.level

    def __getattr__(self, item):
        if hasattr(self.flat, item):
            return getattr(self.flat, item)
        else:
            return None



from typing import List

from pydantic import BaseModel, Field

from .character import CharacterInfo
from .players import PlayerInfo


class EnkaData(BaseModel):
    """All data about the player"""
    player: PlayerInfo = Field({}, alias="playerInfo")
    """Basic player info"""
    characters: List[CharacterInfo] = Field([], alias="avatarInfoList")
    """Player characters shown in game"""
    ttl: int = 0



import setuptools

setuptools.setup(
    name="enka.py",
    version="1.1.5",
    author="pwnless",
    author_email="mystream@riseup.net",
    description="Library for fetching JSON data from site https://enka.network/",
    long_description=open("README.md", "r", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/pwnless/enka.py",
    keywords=['enkapy.py', 'enkapy', 'enka.network', 'genshinapi'],
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    zip_safe=False,
    install_requires=[
        "pydantic",
        "aiohttp",
        "aiocache"
    ],
    python_requires=">=3.6",
    include_package_data=True
)



class EnkaError(Exception):
    pass


class ValidateUIDError(EnkaError):
    pass


class UIDNotFounded(EnkaError):
    pass


class WrongUIDFormat(EnkaError):
    pass


class PlayerDoesNOTExist(EnkaError):
    pass


class GameMaintenance(EnkaError):
    pass


class RateLimited(EnkaError):
    pass


class GeneralServerError(EnkaError):
    pass
